# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

"""Build environment images."""

import argparse
import os
import shutil
import sys
import tempfile
from collections import Counter
from concurrent.futures import as_completed, ThreadPoolExecutor
from datetime import timedelta
from pathlib import Path
from ruamel.yaml import YAML
from subprocess import run, PIPE, STDOUT
from timeit import default_timer as timer
from typing import List, Tuple

import azureml.assets as assets
import azureml.assets.util as util
from azureml.assets.util import logger

TASK_FILENAME = "_acr_build_task.yaml"
BUILD_STEP_TIMEOUT_SECONDS = 60 * 90
SCAN_STEP_TIMEOUT_SECONDS = 60 * 20
DEFAULT_STEP_TIMEOUT_SECONDS = 60 * 30
TRIVY_TIMEOUT = "15m0s"
SUCCESS_COUNT = "success_count"
FAILED_COUNT = "failed_count"
COUNTERS = [SUCCESS_COUNT, FAILED_COUNT]
BUILT_IMAGES = "built_images"


def create_acr_task(image_name: str,
                    dockerfile: str,
                    os: assets.Os,
                    task_filename: str,
                    test_command: str = None,
                    push: bool = False,
                    trivy_url: str = None) -> int:
    """Create ACR build task file.

    Args:
        image_name (str): Image name.
        dockerfile (str): Dockerfile name.
        os (assets.Os): Operating system of the image.
        task_filename (str): File to which the task will be written.
        test_command (str, optional): Command used to test the image. Defaults to None.
        push (bool, optional): Push the image to the ACR. Defaults to False.
        trivy_url (str, optional): URL to download Trivy for vulnerability scanning. Defaults to None.

    Returns:
        int: Cumulative total of step timeouts, to be used when creating the ACR task.
    """
    # Start task with just the build step
    task = {
        'version': 'v1.1.0',
        'stepTimeout': DEFAULT_STEP_TIMEOUT_SECONDS,
        'steps': [{
            'id': "build",
            'timeout': BUILD_STEP_TIMEOUT_SECONDS,
            'build': f"-t $Registry/{image_name} -f {dockerfile} ."
        }]}

    # Add command to output packages
    if os is assets.Os.LINUX:
        # Quoted string required to handle && and ||
        cmd = r'/bin/sh -c "[ -n \"\$CONDA_DEFAULT_ENV\" ] && conda env export || pip freeze"'
    else:
        # Requires Windows batch
        cmd = r'cmd /C "if defined CONDA_DEFAULT_ENV (conda env export) else (pip freeze)"'
    task['steps'].append({
        'id': "package_export",
        'cmd': f"$Registry/{image_name} {cmd}",
        'ignoreErrors': True
    })

    # Add test command if provided
    if test_command:
        task['steps'].append({
            'id': "test",
            'cmd': f"$Registry/{image_name} {test_command}"
        })

    # Add vulnerability scanning step if requested
    if trivy_url is not None:
        if os is assets.Os.LINUX:
            task['steps'].append({
                'id': "scan",
                'timeout': SCAN_STEP_TIMEOUT_SECONDS,
                'cmd': f"aquasec/trivy -q --timeout {TRIVY_TIMEOUT} image --scanners vuln --ignore-unfixed "
                       f"$Registry/{image_name}",
                'ignoreErrors': True,
                'privileged': True
            })
        else:
            logger.log_warning(f"Skipped vulnerability scan for Windows image {image_name} - not supported")

    # Add push step if requested
    if push:
        task['steps'].append({
            'id': "push",
            'push': [f"$Registry/{image_name}"]
        })

    # Write YAML file to disk
    with open(task_filename, "w") as f:
        yaml = YAML()
        yaml.default_flow_style = False
        YAML().dump(task, f)

    # Compute task timeout by adding all step timeouts
    return sum([step.get('timeout', DEFAULT_STEP_TIMEOUT_SECONDS) for step in task['steps']])


def build_image(asset_config: assets.AssetConfig,
                env_config: assets.EnvironmentConfig,
                image_name: str,
                build_log: Path,
                resource_group: str = None,
                registry: str = None,
                test_command: str = None,
                push: bool = False,
                trivy_url: str = None) -> Tuple[assets.AssetConfig, str, int, str]:
    """Build a Docker image locally or via ACR task.

    Args:
        asset_config (assets.AssetConfig): Asset config.
        env_config (assets.EnvironmentConfig): Environment config.
        image_name (str): Image name.
        build_log (Path): File to which the build log will be written.
        resource_group (str, optional): Resource group of the ACR. Defaults to None.
        registry (str, optional): ACR name. Defaults to None.
        test_command (str, optional): Command used to test the image. Defaults to None.
        push (bool, optional): Push the image to the ACR. Defaults to False.
        trivy_url (str, optional): URL to download Trivy for vulnerability scanning. Defaults to None.

    Returns:
        Tuple[assets.AssetConfig, str, int, str]: Asset config, image name, return code, and contents of stdout.
    """
    logger.print(f"Building image for {asset_config.name}")
    start = timer()
    with tempfile.TemporaryDirectory() as temp_dir:
        if registry is not None:
            # Build via ACR task
            temp_dir_path = Path(temp_dir)
            shutil.copytree(env_config.context_dir_with_path, temp_dir_path, dirs_exist_ok=True)
            build_context_dir = temp_dir_path
            task_timeout = create_acr_task(image_name=image_name, dockerfile=env_config.dockerfile,
                                           os=env_config.os, task_filename=build_context_dir / TASK_FILENAME,
                                           test_command=test_command, push=push, trivy_url=trivy_url)
            cmd = ["az", "acr", "run", "-g", resource_group, "-r", registry, "--platform", env_config.os.value,
                   "--timeout", str(task_timeout), "-f", TASK_FILENAME, "."]
        else:
            # Build locally
            build_context_dir = env_config.context_dir_with_path
            cmd = ["docker", "build", "--file", env_config.dockerfile, "--progress", "plain", "--tag", image_name, "."]
        p = run(cmd,
                cwd=build_context_dir,
                stdout=PIPE,
                stderr=STDOUT)
    end = timer()
    logger.print(f"Image for {asset_config.name} built in {timedelta(seconds=end-start)}")
    os.makedirs(build_log.parent, exist_ok=True)
    with open(build_log, "w") as f:
        f.write(p.stdout.decode())
    return (asset_config, image_name, p.returncode, p.stdout.decode())


# Doesn't support ACR yet
def get_image_digest(image_name: str) -> str:
    """Get image digest for a local image.

    Args:
        image_name (str): Image name.

    Returns:
        str: Image digest.
    """
    p = run(["docker", "image", "inspect", image_name, "--format=\"{{index .Id}}\""],
            stdout=PIPE,
            stderr=STDOUT)
    if p.returncode == 0:
        return p.stdout.decode()
    else:
        logger.log_warning(f"Failed to get image digest for {image_name}: {p.stdout.decode()}")
        return None


def build_images(input_dirs: List[Path],
                 asset_config_filename: str,
                 output_directory: Path,
                 build_logs_dir: Path,
                 pin_versions: bool,
                 max_parallel: int,
                 changed_files: List[Path],
                 tag_with_version: bool,
                 os_to_build: assets.Os = None,
                 resource_group: str = None,
                 registry: str = None,
                 test_command: str = None,
                 push: bool = False,
                 use_version_dirs: bool = False,
                 trivy_url: str = None) -> bool:
    """Build Docker images in parallel, either locally or via ACR.

    Args:
        input_dirs (List[Path]): Input directories where environments are located.
        asset_config_filename (str): Asset config filename to search for.
        output_directory (Path): Directory to which assets successfully built will be copied.
        build_logs_dir (Path): Directory to receive build logs.
        pin_versions (bool): Pin image versions.
        max_parallel (int): Maximum number of images to build in parallel.
        changed_files (List[Path]): List of changed files, used to filter environments.
        tag_with_version (bool): Tag image with asset version.
        os_to_build (assets.Os, optional): Operating system to build on via ACR. Defaults to None.
        resource_group (str, optional): Resource group name for ACR builds. Defaults to None.
        registry (str, optional): ACR name. Defaults to None.
        test_command (str, optional): Command used to test images. Defaults to None.
        push (bool, optional): Push images to ACR. Defaults to False.
        use_version_dirs (bool, optional): Use version directories for output. Defaults to False.
        trivy_url (str, optional): URL to download Trivy for vulnerability scanning. Defaults to None.

    Returns:
        bool: True if all images were build successfully, otherwise False.
    """
    counters = Counter()
    with ThreadPoolExecutor(max_parallel) as pool:
        # Find environments under image root directories
        futures = []
        for asset_config in util.find_assets(input_dirs, asset_config_filename, assets.AssetType.ENVIRONMENT,
                                             changed_files):
            env_config = asset_config.extra_config_as_object()

            # Filter by OS
            if os_to_build and env_config.os != os_to_build:
                logger.print(f"Skipping build of image for {asset_config.name}: "
                             f"Operating system {env_config.os.value} != {os_to_build.value}")
                continue

            # Pin versions
            if pin_versions:
                try:
                    assets.pin_env_files(env_config)
                except Exception as e:
                    logger.log_error(f"Failed to pin versions for {asset_config.name}: {e}")
                    counters[FAILED_COUNT] += 1
                    continue

            # Skip environments without build context
            if not env_config.build_enabled:
                logger.print(f"Skipping build of image for {asset_config.name}: No build context specified")

                # Copy file to output directory without building
                if output_directory:
                    util.copy_asset_to_output_dir(asset_config=asset_config, output_directory=output_directory,
                                                  add_subdir=True, use_version_dir=use_version_dirs)
                continue

            # Don't push environments that aren't published
            push_this_image = push and env_config.publish_enabled
            if push and not push_this_image:
                logger.print(f"Skipping push of image for {asset_config.name}: Not published")

            # Tag with version from spec
            if tag_with_version:
                version = asset_config.version
                image_name = env_config.get_image_name_with_tag(version)
            else:
                image_name = env_config.image_name

            # Start building image
            build_log = build_logs_dir / f"{asset_config.name}.log"
            futures.append(pool.submit(build_image, asset_config=asset_config, env_config=env_config,
                                       image_name=image_name, build_log=build_log, resource_group=resource_group,
                                       registry=registry, test_command=test_command, push=push_this_image,
                                       trivy_url=trivy_url))

        # Wait for builds to complete
        built_images = []
        for future in as_completed(futures):
            (asset_config, image_name, return_code, output) = future.result()
            logger.start_group(f"{asset_config.name} build log")
            logger.print(output)
            logger.end_group()
            if return_code != 0:
                logger.log_error(f"Build of image for {asset_config.name} failed with exit status {return_code}",
                                 "Build failure")
                counters[FAILED_COUNT] += 1
            else:
                logger.log_debug(f"Successfully built image for {asset_config.name}")
                counters[SUCCESS_COUNT] += 1
                built_images.append(image_name)
                if output_directory:
                    util.copy_asset_to_output_dir(asset_config=asset_config, output_directory=output_directory,
                                                  add_subdir=True, use_version_dir=use_version_dirs)

    # Set variables
    for counter_name in COUNTERS:
        logger.set_output(counter_name, counters[counter_name])
    logger.set_output(BUILT_IMAGES, ",".join(built_images))

    if counters[FAILED_COUNT] > 0:
        logger.log_error(f"{counters[FAILED_COUNT]} environment image(s) failed to build")
        return False
    return True


if __name__ == '__main__':
    # Handle command-line args
    parser = argparse.ArgumentParser()
    parser.add_argument("-i", "--input-dirs", required=True,
                        help="Comma-separated list of directories containing environments to build")
    parser.add_argument("-a", "--asset-config-filename", default=assets.DEFAULT_ASSET_FILENAME,
                        help="Asset config file name to search for")
    parser.add_argument("-o", "--output-directory", type=Path,
                        help="Directory to which successfully built environments will be written")
    parser.add_argument("-l", "--build-logs-dir", required=True, type=Path,
                        help="Directory to receive build logs")
    parser.add_argument("-p", "--max-parallel", type=int, default=25,
                        help="Maximum number of images to build at the same time")
    parser.add_argument("-c", "--changed-files",
                        help="Comma-separated list of changed files, used to filter images")
    parser.add_argument("-O", "--os-to-build", choices=[i.value for i in list(assets.Os)],
                        help="Only build environments based on this OS")
    parser.add_argument("-r", "--registry",
                        help="Container registry on which to build images")
    parser.add_argument("-g", "--resource-group",
                        help="Resource group containing the container registry")
    parser.add_argument("-P", "--pin-versions", action="store_true",
                        help="Pin images/packages to latest versions")
    parser.add_argument("-t", "--tag-with-version", action="store_true",
                        help="Tag image names using the version in the asset's spec file")
    parser.add_argument("-T", "--test-command",
                        help="If building on ACR, command used to test image, relative to build context root")
    parser.add_argument("-u", "--push", action="store_true",
                        help="If building on ACR, push after building and (optionally) testing")
    parser.add_argument("-v", "--use-version-dirs", action="store_true",
                        help="Use version directories when storing assets in output directory")
    parser.add_argument("-U", "--trivy-url", type=str,
                        help="URL to download Trivy to scan for vulnerabilities")
    args = parser.parse_args()

    # Ensure dependent args are present
    if args.registry and not args.resource_group:
        parser.error("If --registry is specified then --resource-group is also required")
    if args.test_command and not (args.registry and args.resource_group):
        parser.error("--test-command requires both --registry and --resource-group")
    if args.push and not (args.registry and args.resource_group):
        parser.error("--push requires both --registry and --resource-group")
    if args.use_version_dirs and not args.output_directory:
        parser.error("--use-version-dirs requires --output-directory")

    # Reformat arg values
    input_dirs = [Path(d) for d in args.input_dirs.split(",")]
    changed_files = [Path(f) for f in args.changed_files.split(",")] if args.changed_files else []
    os_to_build = assets.Os(args.os_to_build) if args.os_to_build else None

    # Build images
    success = build_images(input_dirs=input_dirs,
                           asset_config_filename=args.asset_config_filename,
                           output_directory=args.output_directory,
                           build_logs_dir=args.build_logs_dir,
                           pin_versions=args.pin_versions,
                           max_parallel=args.max_parallel,
                           changed_files=changed_files,
                           tag_with_version=args.tag_with_version,
                           os_to_build=os_to_build,
                           resource_group=args.resource_group,
                           registry=args.registry,
                           test_command=args.test_command,
                           push=args.push,
                           use_version_dirs=args.use_version_dirs,
                           trivy_url=args.trivy_url)
    if not success:
        sys.exit(1)



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

"""Validate assets."""

import os
import argparse
import json
import re
import sys
from collections import defaultdict
from pathlib import Path
from ruamel.yaml import YAML
from typing import List

from azure.ai.ml import load_model
from azure.ai.ml.entities import Model
from azure.ai.ml.operations._run_history_constants import JobStatus
from azure.core.exceptions import ClientAuthenticationError
from azure.identity import AzureCliCredential

import azureml.assets as assets
import azureml.assets.util as util
from azureml.assets import PublishLocation, PublishVisibility
from azureml.assets.config import ValidationException
from azureml.assets.util import logger
from azureml.assets.util.sku_utils import get_all_sku_details

ERROR_TEMPLATE = "Validation of {file} failed: {error}"
WARNING_TEMPLATE = "Warning during validation of {file}: {warning}"

# Common naming convention
NAMING_CONVENTION_URL = "https://github.com/Azure/azureml-assets/wiki/Asset-naming-convention"
INVALID_STRINGS = [["azureml", "azure"], "aml"]  # Disallow these in any asset name
NON_MODEL_INVALID_STRINGS = ["microsoft"]  # Allow these in model names and other assets related to models
MODEL_RELATED_ASSETS = [assets.AssetType.MODEL, assets.AssetType.EVALUATIONRESULT]
NON_MODEL_NAME_PATTERN = re.compile(r"^[a-z][a-z0-9_.-]{0,254}$")

# Asset config convention
ASSET_CONFIG_URL = "https://github.com/Azure/azureml-assets/wiki/Assets#assetyaml"

# Model naming convention
MODEL_NAME_PATTERN = re.compile(r"^[A-Za-z][A-Za-z0-9_.-]{0,254}$")

# model validations
MODEL_VALIDATION_RESULTS_FOLDER = "validation_results"
VALIDATION_SUMMARY = "results.json"
SUPPORTED_INFERENCE_SKU_FILE_NAME = "config/supported_inference_skus.json"
SUPPORTED_INFERENCE_SKU_FILE_PATH = Path(__file__).parent / SUPPORTED_INFERENCE_SKU_FILE_NAME


class MLFlowModelProperties:
    """Commonly defined model properties."""

    EVALUATION_RECOMMENDED_SKU = "evaluation-recommended-sku"
    EVALUATION_MIN_SKU_SPEC = "evaluation-min-sku-spec"
    FINETUNE_RECOMMENDED_SKU = "finetune-recommended-sku"
    FINETUNE_MIN_SKU_SPEC = "finetune-min-sku-spec"
    INFERENCE_RECOMMENDED_SKU = "inference-recommended-sku"
    INFERENCE_MIN_SKU_SPEC = "inference-min-sku-spec"

    FINETUNING_TASKS = "finetuning-tasks"
    SHARED_COMPUTE_CAPACITY = "SharedComputeCapacityEnabled"


class MLFlowModelTags:
    """Commonly defined model tags."""

    TASK = "task"
    LICENSE = "license"
    AUTHOR = "author"

    EVALUATION_COMPUTE_ALLOWLIST = "evaluation_compute_allow_list"
    FINETUNE_COMPUTE_ALLOWLIST = "finetune_compute_allow_list"
    FINETUNING_DEFAULTS = "model_specific_defaults"
    INFERENCE_COMPUTE_ALLOWLIST = "inference_compute_allow_list"
    INFERENCE_SUPPORTED_ENVS = "inference_supported_envs"

    # This enables model to use shared quota for deployment
    SHARED_COMPUTE_CAPACITY = "SharedComputeCapacityEnabled"

    # make sure all prod models has hiddenlayerscanned tag added
    # model scan and tag has to be added manually currently
    HIDDEN_LAYERS_SCANNED = "hiddenlayerscanned"


class ModelValidationState:
    """Enums for storing validation results state.

    Keeping it consistent with Job state enums. Only the essential ones would be used here.
    """

    NOT_STARTED = JobStatus.NOT_STARTED
    COMPLETED = JobStatus.COMPLETED
    FAILED = JobStatus.FAILED


class ModelValidationOverallSummary:
    """Enum to capture status of all validations."""

    BATCH_DEPLOYMENT = "BatchDeployment"
    ONLINE_DEPLOYMENT = "OnlineDeployment"
    VALIDATION_RUN = "ValidationRun"
    BUILD_URI = "BuildUri"

    @staticmethod
    def get_default_summary():
        """Return model validation default summary."""
        return {
            ModelValidationOverallSummary.BATCH_DEPLOYMENT: ModelValidationState.NOT_STARTED,
            ModelValidationOverallSummary.BATCH_ONLINE_DEPLOYMENTDEPLOYMENT: ModelValidationState.NOT_STARTED,
            ModelValidationOverallSummary.VALIDATION_RUN: ModelValidationState.NOT_STARTED,
        }


# Environment naming convention
ENVIRONMENT_NAME_PATTERN = re.compile(r"^[a-z][a-z0-9.-]{0,254}$")
COMMON_VERSION = r"[0-9]+(?:\.[0-9]+)?(?=-|$)"
FRAMEWORKS = ["pytorch", "sklearn", "tensorflow"]
FRAMEWORK_VERSION = f"(?:{'|'.join(FRAMEWORKS)})-{COMMON_VERSION}"
FRAMEWORK_VERSION_PATTERN = re.compile(FRAMEWORK_VERSION)
INVALID_ENVIRONMENT_STRINGS = ["ubuntu", "cpu"]
OPERATING_SYSTEMS = ["centos", "debian", "win"]
OPERATING_SYSTEM_PATTERN = re.compile(r"(?:centos|debian|\bwin\b)")
OPERATING_SYSTEM_VERSION = f"(?:{'|'.join(OPERATING_SYSTEMS)}){COMMON_VERSION}"
OPERATING_SYSTEM_VERSION_PATTERN = re.compile(OPERATING_SYSTEM_VERSION)
PYTHON_VERSION = r"py3(?:8|9|1[0-9]+)"
PYTHON_VERSION_PATTERN = re.compile(PYTHON_VERSION)
GPU_DRIVERS = ["cuda", "nccl"]
GPU_DRIVER_VERSION = f"(?:{'|'.join(GPU_DRIVERS)}){COMMON_VERSION}"
GPU_DRIVER_VERSION_PATTERN = re.compile(GPU_DRIVER_VERSION)
ENVIRONMENT_NAME_FULL_PATTERN = re.compile("".join([
    FRAMEWORK_VERSION,
    f"(?:-{OPERATING_SYSTEM_VERSION})?",
    f"(?:-{PYTHON_VERSION})?",
    f"(?:-(?:{GPU_DRIVER_VERSION}|gpu))?",
]))

# Dockerfile convention
DOCKERFILE_IMAGE_PATTERN = re.compile(r"^FROM\s+mcr\.microsoft\.com/azureml/curated/", re.IGNORECASE | re.MULTILINE)

# Image naming convention
IMAGE_NAME_PATTERN = re.compile(r"^azureml/curated/(.+)")

# Docker build context stuff to not allow
BUILD_CONTEXT_DISALLOWED_PATTERNS = [
    re.compile(r"extra-index-url", re.IGNORECASE),
]

# Directory path for config files related to asset validation, relative to the current source path
CONFIG_DIRECTORY = "config"


def _log_error(file: Path, error: str) -> None:
    """Log error.

    Args:
        file (Path): File with an issue.
        error (str): Error message.
    """
    logger.log_error(ERROR_TEMPLATE.format(
        file=file.as_posix(),
        error=error
    ))


def _log_warning(file: Path, warning: str) -> None:
    """Log warning.

    Args:
        file (Path): File with an issue.
        warning (str): Warning message.
    """
    logger.log_warning(WARNING_TEMPLATE.format(
        file=file.as_posix(),
        warning=warning
    ))


def validate_environment_name(asset_config: assets.AssetConfig) -> int:
    """Validate environment name.

    Args:
        asset_config (AssetConfig): Asset config.

    Returns:
        int: Number of errors.
    """
    error_count = 0
    asset_name = asset_config.name

    # Check for invalid characters
    if not ENVIRONMENT_NAME_PATTERN.match(asset_name):
        _log_error(asset_config.file_name_with_path, "Name contains invalid characters")
        error_count += 1

    # Check for invalid strings
    for invalid_string in INVALID_ENVIRONMENT_STRINGS:
        if invalid_string in asset_name:
            _log_error(asset_config.file_name_with_path,
                       f"Name '{asset_name}' contains invalid string '{invalid_string}'")
            error_count += 1

    # Check for missing frameworks and version
    frameworks_found = [f for f in FRAMEWORKS if f in asset_name]
    if len(frameworks_found) == 0:
        _log_warning(asset_config.file_name_with_path, f"Name '{asset_name}' is missing framework")
    else:
        # Check framework version
        if not FRAMEWORK_VERSION_PATTERN.search(asset_name):
            _log_error(asset_config.file_name_with_path, f"Name '{asset_name}' is missing framework version")
            error_count += 1

    # Check operating system and version
    if (OPERATING_SYSTEM_PATTERN.search(asset_name) and
            not OPERATING_SYSTEM_VERSION_PATTERN.search(asset_name)):
        _log_error(asset_config.file_name_with_path,
                   f"Name '{asset_name}' is missing operating system version")
        error_count += 1

    # Check python version
    if PYTHON_VERSION_PATTERN.search(asset_name):
        _log_warning(asset_config.file_name_with_path,
                     f"Name '{asset_name}' should only contain Python version if absolutely necessary")

    # Check GPU driver and version
    gpu_drivers_found = [f for f in GPU_DRIVERS if f in asset_name]
    if gpu_drivers_found:
        if not GPU_DRIVER_VERSION_PATTERN.search(asset_name):
            _log_error(asset_config.file_name_with_path, f"Name '{asset_name}' is missing GPU driver version")
            error_count += 1
        if "gpu" in asset_name:
            _log_error(asset_config.file_name_with_path,
                       f"Name '{asset_name}' should not contain both GPU driver and 'gpu'")
            error_count += 1

    # Check for ordering
    if frameworks_found and not ENVIRONMENT_NAME_FULL_PATTERN.search(asset_name):
        _log_error(asset_config.file_name_with_path, f"Name '{asset_name}' elements are out of order")
        error_count += 1

    return error_count


def validate_environment_version(asset_config: assets.AssetConfig) -> int:
    """Validate environment version.

    Args:
        asset_config (AssetConfig): Asset config.

    Returns:
        int: Number of errors.
    """
    if not asset_config.auto_version:
        _log_error(asset_config.file_name_with_path,
                   f"Environment version must be auto but is {asset_config.version}")
        return 1

    return 0


def validate_dockerfile(environment_config: assets.EnvironmentConfig) -> int:
    """Validate Dockerfile.

    Args:
        environment_config (EnvironmentConfig): Environment config.

    Returns:
        int: Number of errors.
    """
    error_count = 0
    dockerfile = environment_config.get_dockerfile_contents()
    dockerfile = dockerfile.replace("\r\n", "\n")

    if DOCKERFILE_IMAGE_PATTERN.search(dockerfile):
        _log_error(environment_config.dockerfile_with_path,
                   "Referencing curated environment images in Dockerfile is not allowed")
        error_count += 1

    return error_count


def validate_build_context(environment_config: assets.EnvironmentConfig) -> int:
    """Validate environment build context.

    Args:
        environment_config (EnvironmentConfig): Environment config.

    Returns:
        int: Number of errors.
    """
    error_count = 0
    # Iterate over all files in the build context
    for file_path in environment_config.release_paths:
        with open(file_path) as f:
            # Read file into memory, normalize EOL characters
            contents = f.read()
            contents = contents.replace("\r\n", "\n")

            # Check disallowed pattersn
            for pattern in BUILD_CONTEXT_DISALLOWED_PATTERNS:
                if pattern.search(contents):
                    _log_error(file_path, f"Found disallowed pattern '{pattern.pattern}'")
                    error_count += 1

    return error_count


def validate_image_publishing(asset_config: assets.AssetConfig,
                              environment_config: assets.EnvironmentConfig = None) -> int:
    """Validate environment image publishing info.

    Args:
        asset_config (AssetConfig): Asset config.
        environment_config (EnvironmentConfig, optional): Environment config. Defaults to None.

    Returns:
        int: Number of errors.
    """
    error_count = 0
    asset_name = asset_config.name
    environment_config = environment_config or asset_config.extra_config_as_object()

    # Check image name
    if (match := IMAGE_NAME_PATTERN.match(environment_config.image_name)) is not None:
        # Ensure image name matches environment name
        if match.group(1) != asset_name:
            _log_error(environment_config.file_name_with_path,
                       f"Image name '{environment_config.image_name}' should be 'azureml/curated/{asset_name}'")
            error_count += 1
    else:
        _log_error(environment_config.file_name_with_path,
                   f"Image name '{environment_config.image_name}' should match pattern azureml/curated/<env_name>")
        error_count += 1

    # Check build context
    if not environment_config.context_dir_with_path.exists():
        _log_error(environment_config.file_name_with_path,
                   f"Build context directory '{environment_config.context_dir}' not found")
        error_count += 1
    if not environment_config.dockerfile_with_path.exists():
        _log_error(environment_config.file_name_with_path,
                   f"Dockerfile '{environment_config.dockerfile}' not found")
        error_count += 1

    # Check publishing info
    if not environment_config.publish_enabled:
        _log_error(environment_config.file_name_with_path, "Image publishing information is not specified")
        error_count += 1

    # Check publishing details
    if environment_config.publish_location != PublishLocation.MCR:
        _log_error(environment_config.file_name_with_path, "Image publishing location should be 'mcr'")
        error_count += 1
    if environment_config.publish_visibility != PublishVisibility.PUBLIC:
        _log_warning(environment_config.file_name_with_path, "Image publishing visibility should be 'public'")

    return error_count


def validate_name(asset_config: assets.AssetConfig) -> int:
    """Validate asset name.

    Args:
        asset_config (AssetConfig): Asset config.

    Returns:
        int: Number of errors.
    """
    error_count = 0
    asset_name = asset_config.name

    # Check against generic naming pattern
    if not ((asset_config.type is assets.AssetType.MODEL and MODEL_NAME_PATTERN.match(asset_name))
            or NON_MODEL_NAME_PATTERN.match(asset_name)):
        _log_error(asset_config.file_name_with_path, f"Name '{asset_name}' contains invalid characters")
        error_count += 1

    # Check for invalid strings
    invalid_strings = INVALID_STRINGS + [asset_config.type.value]
    if asset_config.type not in MODEL_RELATED_ASSETS:
        invalid_strings += [NON_MODEL_INVALID_STRINGS]
    asset_name_lowercase = asset_name.lower()
    for string_group in invalid_strings:
        # Coerce into a list
        string_group_list = string_group if isinstance(string_group, list) else [string_group]

        for invalid_string in string_group_list:
            if invalid_string in asset_name_lowercase:
                # Complain only about the first matching string
                _log_error(asset_config.file_name_with_path,
                           f"Name '{asset_name}' contains invalid string '{invalid_string}'")
                error_count += 1
                break

    # Validate environment name
    if asset_config.type == assets.AssetType.ENVIRONMENT:
        error_count += validate_environment_name(asset_config)

    return error_count


def validate_tests(asset_config: assets.AssetConfig) -> int:
    """Validate pytest information.

    Args:
        asset_config (assets.AssetConfig): Asset config.

    Returns:
        int: Number of errors.
    """
    error_count = 0

    # Bail early if pytest is not enabled
    if not asset_config.pytest_enabled:
        return error_count

    # Check file/dir references
    if not asset_config.pytest_tests_dir_with_path.exists():
        _log_error(asset_config.file_name_with_path,
                   f"pytest.tests_dir directory '{asset_config.pytest_tests_dir}' not found")
        error_count += 1
    if asset_config.pytest_conda_environment and not asset_config.pytest_conda_environment_with_path.exists():
        _log_error(asset_config.file_name_with_path,
                   f"pytest.conda_environment file '{asset_config.pytest_conda_environment}' not found")
        error_count += 1
    if asset_config.pytest_pip_requirements and not asset_config.pytest_pip_requirements_with_path.exists():
        _log_error(asset_config.file_name_with_path,
                   f"pytest.pip_requirements file '{asset_config.pytest_pip_requirements}' not found")
        error_count += 1

    return error_count


def validate_categories(asset_config: assets.AssetConfig) -> int:
    """Validate asset categories.

    Args:
        asset_config (AssetConfig): Asset config.

    Returns:
        int: Number of errors.
    """
    if not asset_config.categories:
        _log_error(asset_config.file_name_with_path, "Categories not found")
        return 1
    return 0


def validate_tags(asset_config: assets.AssetConfig, valid_tags_filename: str) -> int:
    """Validate proper tags for an asset.

    Args:
        asset_config (AssetConfig): Asset config.
        valid_tags_filename (str): Yaml filename in the config directory defining valid tags for the asset.

    Returns:
        int: Number of errors.
    """
    error_count = 0

    with open(Path(__file__).parent / CONFIG_DIRECTORY / valid_tags_filename) as f:
        valid_tags = YAML().load(f)

    asset_spec = asset_config._spec._yaml
    asset_tags = asset_spec.get('tags')

    for tag in valid_tags:
        tag_info = valid_tags[tag]
        if tag_info.get('required') and (not asset_tags or tag not in asset_tags):
            _log_error(asset_config.file_name_with_path,
                       f"Asset '{asset_config.name}' is missing required tag '{tag}'")
            error_count += 1
            continue

        if not asset_tags or tag not in asset_tags:
            continue

        tag_value = asset_tags[tag]
        if not isinstance(tag_value, str):
            _log_error(asset_config.file_name_with_path,
                       f"Asset '{asset_config.name}' has non-string '{type(tag_value)}' for tag '{tag}'")
            error_count += 1
            continue

        valid_tag_values = tag_info.get('values')
        if valid_tag_values:
            if tag_info.get('allow_multiple'):
                tag_values = tag_value.split(',')
                for single_tag_value in tag_values:
                    if single_tag_value not in valid_tag_values:
                        _log_error(asset_config.file_name_with_path,
                                   f"Asset '{asset_config.name}' has invalid value '{single_tag_value}' for tag '{tag}'. Valid values are {valid_tag_values}")  # noqa: E501
                        error_count += 1
                        continue
            else:
                if tag_value not in valid_tag_values:
                    _log_error(asset_config.file_name_with_path,
                               f"Asset '{asset_config.name}' has invalid value '{tag_value}' for tag '{tag}'. Valid values are {valid_tag_values}")  # noqa: E501
                    error_count += 1
                    continue

    return error_count


def confirm_model_validation_results(
    latest_asset_config: assets.AssetConfig,
    validated_asset_config: assets.AssetConfig
) -> int:
    """Compare latest model with validation results.

    Args:
        latest_asset_config (assets.AssetConfig): asset config for latest model
        validated_asset_config (assets.AssetConfig): asset cofig for validated model

    Returns:
        int: Number of errors.

    """
    error_count = 0
    try:
        latest_model_config: assets.ModelConfig = latest_asset_config.extra_config_as_object()
        if latest_model_config.type != assets.config.ModelType.MLFLOW:
            logger.print(
                f"Bypass validation for {latest_asset_config.name} as model type is: {latest_model_config.type.value}"
            )
            return 0

        if not validated_asset_config:
            logger.log_error(f"Validated asset config is None for {latest_asset_config.name}")
            return 1

        validated_model_config: assets.ModelConfig = validated_asset_config.extra_config_as_object()
        logger.print(f"Comparing validated and latest model asset files for {latest_asset_config.name}")

        latest_model_path_uri = latest_model_config.path.uri
        if (latest_model_config.path.type == assets.PathType.AZUREBLOB):
            latest_model_path_uri = latest_model_config.path.uri.split("?")[0]

        validated_model_path_uri = validated_model_config.path.uri
        if (latest_model_config.path.type == assets.PathType.AZUREBLOB):
            validated_model_path_uri = validated_model_config.path.uri.split("?")[0]

        if not (
            latest_model_config.path.type == validated_model_config.path.type and
            latest_model_path_uri == validated_model_path_uri and
            latest_model_config.description == validated_model_config.description and
            latest_model_config.type == validated_model_config.type
        ):
            logger.log_error(
                "Validated model config does not match with latest model. "
                "Either last validation run for model had failed or its still running."
            )

            if latest_model_config.type != validated_model_config.type:
                logger.log_warning(
                    f"latest_model_config_type: [{latest_model_config.type}] and "
                    f"validated_model_config_type: [{validated_model_config.type}] does not match."
                )
                error_count += 1

            if latest_model_config.path.type != validated_model_config.path.type:
                logger.log_warning(
                    f"latest_model_config_path_type: [{latest_model_config.path.type}] and "
                    f"validated_model_config_path_type: [{validated_model_config.path.type}] does not match."
                )
                error_count += 1

            if latest_model_path_uri != validated_model_path_uri:
                logger.log_warning(
                    f"latest_model_config_path_uri: [{latest_model_path_uri}] and "
                    f"validated_model_config_path_uri: [{validated_model_path_uri}] does not match."
                )
                error_count += 1

            if latest_model_config.description != validated_model_config.description:
                logger.log_warning("Description does not match, for latest and validated asset")
                error_count += 1

        # check if spec has changes
        latest_model = load_model(latest_asset_config.spec_with_path)
        validated_model = load_model(validated_asset_config.spec_with_path)

        if latest_model.version != validated_model.version:
            logger.log_error("version mismatch")
            logger.log_warning(f"latest_model tags: [[{latest_model.version}]]")
            logger.log_warning(f"validated_model: [[{validated_model.version}]]")
            error_count += 1

        if latest_model.tags != validated_model.tags:
            logger.log_error("tags mismatch")
            logger.log_warning(f"latest_model tags: [{latest_model.tags}]")
            logger.log_warning(f"validated_model: [{validated_model.tags}]")
            error_count += 1

        if latest_model.properties != validated_model.properties:
            logger.log_error("properties mismatch")
            logger.log_warning(f"latest_model properties: [{latest_model.properties}]")
            logger.log_warning(f"validated_model properties: [{validated_model.properties}]")
            error_count += 1

        if latest_model.description != validated_model.description:
            logger.log_error("description mismatch")
            logger.log_warning(f"latest_model description: [{latest_model.description}]")
            logger.log_warning(f"validated_model description: [{validated_model.description}]")
            error_count += 1

        # check validation results now
        validation_results_dir = validated_asset_config.file_path / MODEL_VALIDATION_RESULTS_FOLDER
        validation_job_details_path = validation_results_dir / VALIDATION_SUMMARY
        if not validation_job_details_path.exists():
            logger.log_error(
                f"{VALIDATION_SUMMARY} missing for model {latest_asset_config.name}. "
                "Either last validation run for model had failed or its still running."
            )
            error_count += 1
        else:
            overall_summary = {}
            with open(validation_job_details_path) as f:
                overall_summary = json.load(f)

            validation_run_status = overall_summary.get(
                ModelValidationOverallSummary.VALIDATION_RUN, ModelValidationState.NOT_STARTED)
            batch_deployment_status = overall_summary.get(
                ModelValidationOverallSummary.BATCH_DEPLOYMENT, ModelValidationState.NOT_STARTED)
            online_deployment_status = overall_summary.get(
                ModelValidationOverallSummary.ONLINE_DEPLOYMENT, ModelValidationState.NOT_STARTED)
            buildUri = overall_summary.get(ModelValidationOverallSummary.BUILD_URI, None)

            if validation_run_status != ModelValidationState.COMPLETED:
                logger.log_error(
                    f"run status for model {latest_asset_config.name} is {validation_run_status}. "
                    "Please ensure that there is a completed model validation job."
                )
                error_count += 1

            # check is batch supported?
            disable_batch_str = latest_model.tags.get("disable-batch", "false")
            supports_batch = True if disable_batch_str.lower() == "false" else False

            if supports_batch and batch_deployment_status != ModelValidationState.COMPLETED:
                logger.log_error(
                    f"batch deployment is supported for {latest_model.name}. "
                    f"But its status is {batch_deployment_status}. "
                    "Please ensure that that batch is validated for the model."
                )
                error_count += 1

            # check if online inference is supported
            supports_inference = (
                True
                if latest_model.tags.get(MLFlowModelTags.INFERENCE_COMPUTE_ALLOWLIST, None)
                or latest_model.properties.get(MLFlowModelProperties.INFERENCE_RECOMMENDED_SKU)
                else False
            )

            if supports_inference and online_deployment_status != ModelValidationState.COMPLETED:
                logger.log_error(
                    f"Online deployment is supported for {latest_model.name}. "
                    f"But its status is {online_deployment_status}. "
                    "Please ensure that that online inference is validated for the model."
                )
                error_count += 1

            if error_count > 0 and buildUri:
                logger.print(f"Check model: {latest_model.name} validation logs here: {buildUri}")

        return error_count
    except Exception as e:
        logger.log_error(
            f"Exception when confirming validation results for model {latest_asset_config.name}. Exception {e}"
        )
        return 1


def validate_model_scenario(
    asset_file_name_with_path: Path,
    model: Model,
    min_sku_prop_name: str,
    recommended_skus_prop_name: str,
    compute_allowlist_tags_name: str,
) -> int:
    """Validate model properties, tags for different scenarios.

    Args:
        asset_file_name_with_path (Path): file path to model asset
        model (Model): model loaded from spec
        min_sku_prop_name (str): min sku property name for the scenario
        recommended_skus_prop_name (str): recommended sku property name for the scenario
        compute_allowlist_tags_name (str): compute allowlist tag name for the scenario

    Returns:
        int: Number of errors.

    """
    error_count = 0
    min_sku = model.properties.get(min_sku_prop_name, "").strip()
    recommended_skus = model.properties.get(recommended_skus_prop_name, "").strip()
    compute_allowlists = set(model.tags.get(compute_allowlist_tags_name, []))

    if not min_sku:
        _log_error(asset_file_name_with_path, f"{min_sku_prop_name} is missing in model properties")
        error_count += 1

    if not recommended_skus:
        _log_error(asset_file_name_with_path, f"{recommended_skus_prop_name} is missing in model properties")
        error_count += 1

    if not compute_allowlists:
        _log_error(asset_file_name_with_path, f"{compute_allowlist_tags_name} is missing in model tags")
        error_count += 1

    recommended_skus = set([sku.strip() for sku in recommended_skus.split(",")])
    if (recommended_skus != compute_allowlists):
        a_minus_b = recommended_skus - compute_allowlists
        b_minus_a = compute_allowlists - recommended_skus
        _log_error(
            asset_file_name_with_path,
            f"{recommended_skus_prop_name} and {compute_allowlist_tags_name} does not match for model.\n"
            f"skus in recommended sku and not in compute allowlists => {a_minus_b}\n"
            f"skus in compute allowlists and not in recommended sku => {b_minus_a}\n"
        )
        error_count += 1

    # confirm min_sku_spec with list of supported computes
    error_count += confirm_min_sku_spec(asset_file_name_with_path, min_sku_prop_name, compute_allowlists, min_sku)

    return error_count


def confirm_min_sku_spec(
    asset_file_name_with_path: Path,
    min_sku_prop_name: str,
    supported_skus: set,
    min_sku_spec: str
):
    """Validate model properties, tags for different scenarios.

    Args:
        asset_file_name_with_path (Path): file path to model asset
        min_sku_prop_name (str): min sku property name for the scenario
        supported_skus (List): supported SKUs for the scenario
        min_sku_spec (str): Scenario min SKU spec

    Returns:
        int: Number of errors.
    """
    subscription_id = os.getenv("SUBSCRIPTION_ID")
    if not subscription_id:
        logger.log_warning("subscription_id missing. Skipping min sku valdn")
        return 0

    try:
        all_sku_details = get_all_sku_details(AzureCliCredential(), subscription_id)
        min_disk = min_cpu_mem = min_ngpus = min_ncpus = -1
        for sku in supported_skus:
            sku_details = all_sku_details.get(sku)
            if not sku_details:
                raise Exception(
                    f"Caught exception while checking {min_sku_prop_name}."
                    f" Either invalid sku {sku} or issue with fetching sku details"
                )

            num_cpus = sku_details["vCPUs"]
            num_gpus = sku_details["gpus"]
            cpu_mem = int(sku_details["memoryGB"])
            disk_space = int(sku_details["maxResourceVolumeMB"] / 1024)

            min_ncpus = min(num_cpus, min_ncpus) if min_ncpus > 0 else num_cpus
            min_ngpus = min(num_gpus, min_ngpus) if min_ngpus >= 0 else num_gpus
            min_cpu_mem = min(cpu_mem, min_cpu_mem) if min_cpu_mem > 0 else cpu_mem
            min_disk = min(disk_space, min_disk) if min_disk > 0 else disk_space

        ncpus, ngpus, mem, disk = [int(item) for item in min_sku_spec.split("|")]
        if ncpus != min_ncpus or ngpus != min_ngpus or mem != min_cpu_mem or disk != min_disk:
            _log_error(
                asset_file_name_with_path,
                f"for {min_sku_prop_name} => "
                f"{ncpus}|{ngpus}|{mem}|{disk} != {min_ncpus}|{min_ngpus}|{min_cpu_mem}|{min_disk}"
            )

            # list of skus larger than current specific min-sku
            skus_failing_valdn = []
            for sku in supported_skus:
                sku_details = all_sku_details.get(sku)
                num_cpus = sku_details["vCPUs"]
                num_gpus = sku_details["gpus"]
                cpu_mem = int(sku_details["memoryGB"])
                disk_space = int(sku_details["maxResourceVolumeMB"] / 1024)

                if num_cpus < ncpus or num_gpus < ngpus or cpu_mem < mem or disk_space < disk:
                    sku_spec = "|".join([str(num_cpus), str(num_gpus), str(cpu_mem), str(disk_space)])
                    skus_failing_valdn.append(f"{sku}: {sku_spec}")

            _log_error(
                asset_file_name_with_path,
                f"for {min_sku_prop_name} => "
                f"SKUs having smaller spec: {skus_failing_valdn}"
            )

            return 1
    except ClientAuthenticationError as e:
        logger.log_warning(f"Failed to log in to Azure, skipping min sku valdn. {e}")
    except Exception as e:
        _log_error(asset_file_name_with_path, f"Exception in fetching SKU details: {e}")
        return 1
    return 0


def validate_model_spec(asset_config: assets.AssetConfig) -> int:
    """Validate model spec.

    Args:
        asset_config (assets.AssetConfig): asset config for model spec

    Returns:
        int: error count
    """
    error_count = 0
    model = model_config = None

    try:
        model = load_model(asset_config.spec_with_path)
    except Exception:
        _log_error(asset_config.file_name_with_path, "Invalid spec file")
        return 1

    try:
        model_config: assets.ModelConfig = asset_config.extra_config_as_object()
    except Exception:
        _log_error(asset_config.file_name_with_path, "Invalid model config")
        return 1

    if model_config.type != assets.config.ModelType.MLFLOW:
        logger.print(
            f"Bypass validation for {asset_config.name} as model type is: {model_config.type.value}"
        )
        return 0

    # confirm must have
    if not model.tags.get(MLFlowModelTags.TASK):
        _log_error(asset_config.file_name_with_path, f"{MLFlowModelTags.TASK} missing")
        error_count += 1

    if not model.tags.get(MLFlowModelTags.LICENSE):
        _log_error(asset_config.file_name_with_path, f"{MLFlowModelTags.LICENSE} missing")
        error_count += 1

    # TODO: skip hidden layer valdn till most models are scanned
    # if MLFlowModelTags.HIDDEN_LAYERS_SCANNED not in model.tags:
    #     _log_error(
    #         asset_config.file_name_with_path,
    #         (
    #             "`hiddenlayerscanned` tag missing. "
    #             "Model is not scanned by HiddenLayer ModelScanner tool. Please scan the model and retry"
    #         )
    #     )
    #     error_count += 1

    # shared compute check
    if MLFlowModelTags.SHARED_COMPUTE_CAPACITY not in model.tags:
        _log_error(asset_config.file_name_with_path, f"Tag {MLFlowModelTags.SHARED_COMPUTE_CAPACITY} missing")
        error_count += 1

    if not model.properties.get(MLFlowModelProperties.SHARED_COMPUTE_CAPACITY, False):
        _log_error(
            asset_config.file_name_with_path, f"Property {MLFlowModelProperties.SHARED_COMPUTE_CAPACITY} is not set"
        )
        error_count += 1

    # if any of the relevant tags or properies is present
    # assume support as true and then fail in valdn
    supports_eval = (
        MLFlowModelTags.EVALUATION_COMPUTE_ALLOWLIST in model.tags
        or MLFlowModelProperties.EVALUATION_RECOMMENDED_SKU in model.properties
        or MLFlowModelProperties.EVALUATION_MIN_SKU_SPEC in model.properties
    )

    # If any of the relevant tags or properies is present
    # assume support as true and then fail in valdn
    supports_ft = (
        MLFlowModelTags.FINETUNE_COMPUTE_ALLOWLIST in model.tags
        or MLFlowModelProperties.FINETUNE_RECOMMENDED_SKU in model.properties
        or MLFlowModelProperties.FINETUNE_MIN_SKU_SPEC in model.properties
        or MLFlowModelProperties.FINETUNING_TASKS in model.properties
    )

    # validate inference compute req.
    error_count += validate_model_scenario(
        asset_config.file_name_with_path,
        model,
        MLFlowModelProperties.INFERENCE_MIN_SKU_SPEC,
        MLFlowModelProperties.INFERENCE_RECOMMENDED_SKU,
        MLFlowModelTags.INFERENCE_COMPUTE_ALLOWLIST,
    )

    # check valid computes for inference
    with open(SUPPORTED_INFERENCE_SKU_FILE_PATH) as f:
        supported_inference_skus = set(json.load(f))
        unsupported_skus_in_spec = [
            sku
            for sku in model.tags.get(MLFlowModelTags.INFERENCE_COMPUTE_ALLOWLIST, [])
            if sku not in supported_inference_skus
        ]
        if unsupported_skus_in_spec:
            _log_error(asset_config.file_name_with_path,
                       f"Unsupported inference SKU in spec: {unsupported_skus_in_spec}")
            error_count += 1

    if supports_eval:
        error_count += validate_model_scenario(
            asset_config.file_name_with_path,
            model,
            MLFlowModelProperties.EVALUATION_MIN_SKU_SPEC,
            MLFlowModelProperties.EVALUATION_RECOMMENDED_SKU,
            MLFlowModelTags.EVALUATION_COMPUTE_ALLOWLIST,
        )

    if supports_ft:
        if not model.properties.get(MLFlowModelProperties.FINETUNING_TASKS):
            _log_error(
                asset_config.file_name_with_path,
                f"{MLFlowModelProperties.FINETUNING_TASKS} not set for supporting finetuning scenario"
            )
            error_count += 1

        error_count += validate_model_scenario(
            asset_config.file_name_with_path,
            model,
            MLFlowModelProperties.FINETUNE_MIN_SKU_SPEC,
            MLFlowModelProperties.FINETUNE_RECOMMENDED_SKU,
            MLFlowModelTags.FINETUNE_COMPUTE_ALLOWLIST,
        )

    return error_count


def get_validated_models_assets_map(model_validation_results_dir: str):
    """Return model assets map."""
    try:
        if not model_validation_results_dir:
            logger.log_warning(
                "Unexpected !!! model_validation_results_dir is None. Model assets might fail in validation."
            )
            return {}

        validated_model_assets: List[assets.AssetConfig] = util.find_assets(
            input_dirs=[Path(model_validation_results_dir)],
            asset_config_filename=assets.DEFAULT_ASSET_FILENAME,
            types=[assets.config.AssetType.MODEL]
        )

        return {model_asset.name: model_asset for model_asset in validated_model_assets}
    except Exception as e:
        logger.log_error(f"Error in creating validated model map => {e}")
        return {}


def validate_assets(input_dirs: List[Path],
                    asset_config_filename: str,
                    model_validation_results_dir: str = None,
                    changed_files: List[Path] = None,
                    check_names: bool = False,
                    check_names_skip_pattern: re.Pattern = None,
                    check_images: bool = False,
                    check_categories: bool = False,
                    check_build_context: bool = False,
                    check_tests: bool = False,
                    check_environment_version: bool = False) -> bool:
    """Validate assets.

    Args:
        input_dirs (List[Path]): Directories containing assets.
        asset_config_filename (str): Asset config filename to search for.
        model_validation_results_dir (str, optional): Dir containing model validation results
        changed_files (List[Path], optional): List of changed files, used to filter assets. Defaults to None.
        check_names (bool, optional): Whether to check asset names. Defaults to False.
        check_names_skip_pattern (re.Pattern, optional): Regex pattern to skip name validation. Defaults to None.
        check_images (bool, optional): Whether to check image names. Defaults to False.
        check_categories (bool, optional): Whether to check asset categories. Defaults to False.
        check_build_context (bool, optional): Whether to check environment build context. Defaults to False.
        check_tests (bool, optional): Whether to check test references. Defaults to False.
        check_environment_version (bool, optional): Whether to check environment version. Defaults to False.

    Raises:
        ValidationException: If validation fails.

    Returns:
        bool: True if assets were successfully validated, otherwise False.
    """
    # Gather list of just changed assets, for later filtering
    changed_assets = util.find_asset_config_files(input_dirs, asset_config_filename, changed_files) if changed_files else None  # noqa: E501
    validated_model_map = get_validated_models_assets_map(model_validation_results_dir)

    # Find assets under input dirs
    asset_count = 0
    error_count = 0
    asset_dirs = defaultdict(list)
    image_names = defaultdict(list)
    for asset_config_path in util.find_asset_config_files(input_dirs, asset_config_filename):
        asset_count += 1
        # Errors only "count" if changed_files was None or the asset was changed
        validate_this = changed_assets is None or asset_config_path in changed_assets

        # Load config
        try:
            asset_config = assets.AssetConfig(asset_config_path)
        except Exception as e:
            if validate_this:
                _log_error(asset_config_path, e)
                error_count += 1
            else:
                _log_warning(asset_config_path, e)
            continue

        # Populate dictionary of asset names to asset config paths
        asset_dirs[f"{asset_config.type.value} {asset_config.name}"].append(asset_config_path)

        # validated_model_map would be empty for non-drop scenario
        if asset_config.type == assets.AssetType.MODEL:
            error_count += validate_model_spec(asset_config)
            # should run during drop creation only
            if validated_model_map:
                error_count += confirm_model_validation_results(
                    asset_config,
                    validated_model_map.get(asset_config.name, None)
                )

        # Populate dictionary of image names to asset config paths
        environment_config = None
        if asset_config.type == assets.AssetType.ENVIRONMENT:
            try:
                environment_config = asset_config.extra_config_as_object()

                # Store fully qualified image name
                image_name = environment_config.image_name
                if environment_config.publish_location:
                    image_name = f"{environment_config.publish_location.value}/{image_name}"
                image_names[image_name].append(asset_config.file_path)
            except Exception as e:
                if validate_this:
                    _log_error(environment_config.file_name_with_path, e)
                    error_count += 1
                else:
                    _log_warning(environment_config.file_name_with_path, e)

        # Checks for changed assets only, or all assets if changed_files was None
        if validate_this:
            # Validate name
            if check_names:
                if check_names_skip_pattern is None or not check_names_skip_pattern.fullmatch(asset_config.full_name):
                    error_count += validate_name(asset_config)
                else:
                    logger.log_debug(f"Skipping name validation for {asset_config.full_name}")

            # Validate pytest information
            if check_tests:
                error_count += validate_tests(asset_config)

            if asset_config.type == assets.AssetType.ENVIRONMENT:
                # Validate Dockerfile
                error_count += validate_dockerfile(asset_config.extra_config_as_object())
                if check_build_context:
                    error_count += validate_build_context(asset_config.extra_config_as_object())

                # Validate environment version
                if check_environment_version:
                    error_count += validate_environment_version(asset_config)

            if asset_config.type == assets.AssetType.EVALUATIONRESULT:
                error_count += validate_tags(asset_config, 'evaluationresult/tag_values_shared.yaml')

                asset_spec = asset_config._spec._yaml
                evaluation_type = asset_spec.get('tags', {}).get('evaluation_type', None)

                if evaluation_type == 'text_generation':
                    error_count += validate_tags(asset_config, 'evaluationresult/tag_values_text_generation.yaml')
                elif evaluation_type == 'text_embeddings':
                    error_count += validate_tags(asset_config, 'evaluationresult/tag_values_text_embeddings.yaml')
                else:
                    _log_error(
                        asset_config.file_name_with_path,
                        f"Asset '{asset_config.name}' has unknown evaluation_type: '{evaluation_type}'"
                    )
                    error_count += 1

            if asset_config.type == assets.AssetType.PROMPT:
                error_count += validate_tags(asset_config, 'tag_values_shared.yaml')
                error_count += validate_tags(asset_config, 'tag_values_prompt.yaml')

            # Validate categories
            if check_categories:
                error_count += validate_categories(asset_config)

            # Validate specific asset types
            if environment_config is not None:
                if check_images:
                    # Check image name
                    error_count += validate_image_publishing(asset_config, environment_config)

            # Validate spec
            try:
                spec = asset_config.spec_as_object()

                # Ensure name and version aren't inconsistent
                if not assets.Config._contains_template(spec.name) and asset_config.name != spec.name:
                    raise ValidationException(f"Asset and spec name mismatch: {asset_config.name} != {spec.name}")
                if not assets.Config._contains_template(spec.version) and asset_config.version != spec.version:
                    raise ValidationException(f"Asset and spec version mismatch: {asset_config.version} != {spec.version}")  # noqa: E501
            except Exception as e:
                _log_error(asset_config.spec_with_path, e)
                error_count += 1

    # Ensure unique assets
    for type_and_name, dirs in asset_dirs.items():
        if len(dirs) > 1:
            dirs_str = [d.as_posix() for d in dirs]
            logger.log_error(f"{type_and_name} found in multiple asset YAMLs: {dirs_str}")
            error_count += 1

    # Ensure unique image names
    for image_name, dirs in image_names.items():
        if len(dirs) > 1:
            dirs_str = [d.as_posix() for d in dirs]
            logger.log_error(f"{image_name} found in multiple assets: {dirs_str}")
            error_count += 1

    logger.print(f"Found {error_count} error(s) in {asset_count} asset(s)")
    return error_count == 0


if __name__ == '__main__':
    # Handle command-line args
    parser = argparse.ArgumentParser()
    parser.add_argument("-i", "--input-dirs", required=True,
                        help="Comma-separated list of directories containing assets")
    parser.add_argument("-m", "--model-validation-results-dir", required=False,
                        help="Dir containing model validation results and validated spec.")
    parser.add_argument("-a", "--asset-config-filename", default=assets.DEFAULT_ASSET_FILENAME,
                        help="Asset config file name to search for")
    parser.add_argument("-c", "--changed-files",
                        help="Comma-separated list of changed files, used to filter assets")
    parser.add_argument("-n", "--check-names", action="store_true",
                        help="Check asset names")
    parser.add_argument("-I", "--check-images", action="store_true",
                        help="Check environment images")
    parser.add_argument("-C", "--check-categories", action="store_true",
                        help="Check asset categories")
    parser.add_argument("-N", "--check-names-skip-pattern", type=re.compile,
                        help="Regex pattern to skip name validation, in the format <type>/<name>/<version>")
    parser.add_argument("-b", "--check-build-context", action="store_true",
                        help="Check environment build context")
    parser.add_argument("-t", "--check-tests", action="store_true",
                        help="Check test references")
    parser.add_argument("-e", "--check-environment-version", action="store_true",
                        help="Check environment version")
    args = parser.parse_args()

    # Convert comma-separated values to lists
    input_dirs = [Path(d) for d in args.input_dirs.split(",")]
    changed_files = [Path(f) for f in args.changed_files.split(",")] if args.changed_files else []

    # Share asset naming convention URL
    if args.check_names:
        logger.print(f"Asset naming convention: {NAMING_CONVENTION_URL}")

    # Share asset config reference
    if args.check_categories:
        logger.print(f"Asset config reference: {ASSET_CONFIG_URL}")

    # Validate assets
    success = validate_assets(input_dirs=input_dirs,
                              asset_config_filename=args.asset_config_filename,
                              changed_files=changed_files,
                              check_names=args.check_names,
                              check_names_skip_pattern=args.check_names_skip_pattern,
                              check_images=args.check_images,
                              check_categories=args.check_categories,
                              check_build_context=args.check_build_context,
                              model_validation_results_dir=args.model_validation_results_dir,
                              check_tests=args.check_tests,
                              check_environment_version=args.check_environment_version)
    if not success:
        sys.exit(1)



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

from .config import (
    AssetConfig,
    AssetType,
    ComponentType,
    Config,
    DEFAULT_ASSET_FILENAME,
    ModelConfig,
    ModelFlavor,
    ModelTaskName,
    ModelType,
    EnvironmentConfig,
    Os,
    PathType,
    PublishLocation,
    PublishVisibility,
    Spec,
)
from .deployment_config import (
    DeploymentConfig,
)
from .update_assets import (
    pin_env_files,
    release_tag_exists,
    get_latest_release_tag_version,
    update_asset,
    update_assets,
)
from .environment.pin_image_versions import get_manifest
from .get_tokens import get_tokens
from .publish_utils import create_asset
from .update_spec import create_template_data, update as update_spec
from .validate_assets import validate_assets
from .validate_tree import validate_tree



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

"""Extract selected assets from release branch into output directory."""

import argparse
import re
from collections import defaultdict
from git import Repo
from pathlib import Path

import azureml.assets as assets
import azureml.assets.util as util
from azureml.assets.util import logger

MATCHED_COUNT = "matched_count"
EXTRACTED_COUNT = "extracted_count"


def extract_tag_released_assets(release_directory_root: Path,
                                output_directory_root: Path,
                                pattern: re.Pattern = None,
                                include_deprecated: bool = False):
    """Extract selected assets from release branch and copy into output directory.

    Args:
        release_directory_root (Path): Release directory location.
        output_directory_root (Path): Output directory.
        pattern (re.Pattern, optional): Regex pattern for assets to extract and copy. Defaults to None.
        include_deprecated (bool, optional): Include deprecated assets. Defaults to False.
    """
    # Gather list of asset in release branch HEAD
    asset_list = set()
    if not include_deprecated:
        for asset_config in util.find_assets(input_dirs=release_directory_root):
            # Store as type/name
            asset_list.add(asset_config.partial_name)

    # Initialize counters
    matched_count = 0
    extracted_count = 0

    # Select tags
    repo = Repo(release_directory_root)
    commits_tags = defaultdict(list)
    for tag in repo.tags:
        if not pattern or pattern.fullmatch(tag.name):
            # Skip deprecated assets, if specified
            if not include_deprecated:
                type, name, _ = assets.AssetConfig.parse_full_name(tag.name)
                partial_name = f"{type.value}/{name}"
                if partial_name not in asset_list:
                    continue
            commits_tags[tag.commit].append(tag)
            matched_count += 1

    # Order commits for efficient checkout
    ordered_commits = sorted(commits_tags.keys(), key=lambda c: c.authored_datetime)

    # Checkout tags
    for commit in ordered_commits:
        tags = commits_tags[commit]
        logger.print(f"Checking out commit {commit} for {len(tags)} tag(s)")
        repo.git().checkout(commit)

        # Iterate over tags
        for tag in tags:
            # Copy asset to output directory
            type, name, version = assets.AssetConfig.parse_full_name(tag.name)
            release_dir = util.get_asset_release_dir_from_parts(type, name, release_directory_root)
            output_directory = util.get_asset_output_dir_from_parts(type, name, output_directory_root, version)
            if release_dir.exists():
                logger.print(f"Copying {type.value} {name}:{version} to {output_directory}")
                util.copy_replace_dir(release_dir, output_directory)
                extracted_count += 1
            else:
                logger.log_warning(f"{type.value.capitalize()} {name}:{version} not found in commit")

    # Reset to release branch
    repo.git().checkout("release")

    # Set variables
    logger.set_output(MATCHED_COUNT, matched_count)
    logger.set_output(EXTRACTED_COUNT, extracted_count)


if __name__ == "__main__":
    # Handle command-line args
    parser = argparse.ArgumentParser()
    parser.add_argument("-r", "--release-directory", required=True, type=Path,
                        help="Directory to which the release branch has been cloned")
    parser.add_argument("-o", "--output-directory", required=True, type=Path,
                        help="Directory to which unreleased assets will be written")
    parser.add_argument("-t", "--pattern", type=re.compile,
                        help="Regex pattern to select assets to extract, in the format <type>/<name>/<version>")
    parser.add_argument("-d", "--include-deprecated", action="store_true",
                        help="Include deprecated assets")
    args = parser.parse_args()

    extract_tag_released_assets(release_directory_root=args.release_directory,
                                output_directory_root=args.output_directory,
                                pattern=args.pattern,
                                include_deprecated=args.include_deprecated)



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

"""Utility functions used to publish assets."""

import json
import re
import shutil
import sys
import azureml.assets.util as util
from pathlib import Path
from string import Template
from subprocess import CompletedProcess, run
from tempfile import TemporaryDirectory
from typing import Dict, List, Tuple, Union
from azureml.assets.config import AssetConfig, AssetType, ComponentType, ModelConfig
from azureml.assets.deployment_config import AssetVersionUpdate
from azureml.assets.model.model_utils import CopyUpdater, prepare_model, update_model_metadata
from azureml.assets.util import logger
from azureml.assets.util.util import resolve_from_file_for_asset
from azure.ai.ml import MLClient
from azure.ai.ml.entities import Component, Environment, Model
from ruamel.yaml import YAML


PROD_SYSTEM_REGISTRY = "azureml"
WORKSPACE_ASSET_PATTERN = re.compile(r"^(?:azureml:)?(.+)(?::(.+)|@(.+))$")
REGISTRY_ENV_PATTERN = re.compile(r"^azureml://registries/(.+)/environments/(.+)/(?:versions/(.+)|labels/(.+))")
REGISTRY_ASSET_TEMPLATE = Template("^azureml://registries/(.+)/$asset_type/(.+)/(?:versions/(.+)|labels/(.+))")
BEARER = r"Bearer.*"
LATEST_LABEL = "latest"


def pluralize_asset_type(asset_type: Union[AssetType, str]) -> str:
    """Return pluralized asset type."""
    # Convert to string if enum
    if isinstance(asset_type, AssetType):
        asset_type = asset_type.value
    return f"{asset_type}s" if asset_type != "data" else asset_type


def sanitize_output(input: str) -> str:
    """Return sanitized string."""
    # Remove sensitive token
    sanitized_output = re.sub(BEARER, "", input)
    return sanitized_output


def update_spec(asset: Union[Component, Environment, Model], spec_path: Path) -> bool:
    """Update the yaml spec file with updated properties in asset.

    :param asset: Asset loaded using load_*(component, environemnt, model) method.
    :type asset: Union[Component, Environment, Model]
    :param spec_path: path to asset spec file
    :type spec_path: Path
    :return: True if spec was successfully updated
    :rtype: bool
    """
    try:
        asset_dict = json.loads(json.dumps(asset._to_dict()))
        util.dump_yaml(asset_dict, spec_path)
        return True
    except Exception as e:
        logger.log_error(f"Failed to update spec: {e}")
    return False


def prepare_model_for_registration(
    model_config: ModelConfig,
    spec_file_path: Path,
    temp_dir: Path,
    ml_client: MLClient,
    copy_updater: CopyUpdater = None,
) -> bool:
    """Prepare model.

    :param model_config: Model Config object
    :type model_config: ModelConfig
    :param spec_file_path: path to model spec file
    :type spec_file_path: Path
    :param temp_dir: temp dir for model operation
    :type temp_dir: Path
    :param ml_client: MLClient object
    :type ml_client: MLClient
    :param copy_updater: CopyUpdater object to update files during azcopy
    :type copy_updater: CopyUpdater
    :return: Model successfully prepared for creation in registry.
    :rtype: bool
    """
    model, success = prepare_model(
        spec_path=spec_file_path, model_config=model_config, temp_dir=temp_dir, ml_client=ml_client,
        copy_updater=copy_updater
    )
    if success:
        success = update_spec(model, spec_file_path)
        logger.print(f"updated spec file? {success}")
    return success


def validate_and_prepare_pipeline_component(
    spec_path: Path,
    registry_name: str,
    version_template: str = None
) -> bool:
    """Validate and update pipeline component spec.

    :param spec_path: Path of loaded component
    :type spec_path: Path
    :param registry_name: name of the registry to create component in
    :type registry_name: str
    :param version_template: version template
    :type version_suffix: str
    :return: True for successful validation and update
    :rtype: bool
    """
    with open(spec_path) as f:
        try:
            pipeline_dict = YAML().load(f)
        except Exception:
            logger.log_error(f"Error in loading component spec at {spec_path}")
            return False

    jobs = pipeline_dict['jobs']
    logger.print(f"Preparing pipeline component {pipeline_dict['name']}")
    updated_jobs = {}

    for job_name, job_details in jobs.items():
        logger.print(f"job {job_name}")
        if not job_details.get('component'):
            # if-else or inline component
            logger.print(f"component not defined for job {job_name}")
            updated_jobs[job_name] = job_details
            continue

        try:
            name, version, label, registry = get_parsed_details_from_asset_uri(
                AssetType.COMPONENT.value, job_details['component'])
        except Exception as e:
            logger.log_error(e)
            return False

        logger.print(
            "Parsed component asset URI details:\n"
            + f"name: {name}\n"
            + f"version: {version}\n"
            + f"label: {label}\n"
            + f"registry: {registry}"
        )

        if registry and registry not in [PROD_SYSTEM_REGISTRY, registry_name]:
            logger.log_warning(
                f"Dependencies should exist in '{registry_name}' or '{PROD_SYSTEM_REGISTRY}'. "
                f"The URI for component '{name}' references registry '{registry}', "
                "and publishing will fail if the release process does not have read access to it."
            )

        # If workspace asset URI is used, use registry we're creating the component in
        if not registry:
            logger.print(f"Workspace asset URI was used, using component from registry {registry}")
            registry = registry_name

        # Check if component's env exists
        final_version = util.apply_version_template(version, version_template)
        asset_details = None
        for ver in [version, final_version]:
            if (asset_details := get_asset_details(
                AssetType.COMPONENT.value, name, ver, registry
            )) is not None:
                break

        if not asset_details:
            logger.log_warning(
                f"dependent component {name} with version {version} not found in registry {registry}"
            )
            return False

        updated_jobs[job_name] = job_details
        updated_jobs[job_name]['component'] = asset_details["id"]

    pipeline_dict['jobs'] = updated_jobs

    try:
        util.dump_yaml(pipeline_dict, spec_path)
    except Exception:
        logger.log_error(f"Component update failed for asset spec path: {spec_path}")
        return False
    return True


def get_environment_asset_id(
    environment_id: str,
    registry_name: str,
    version_template: str = None
) -> Union[object, None]:
    """Convert an environment reference into a full asset ID.

    :param environment_id: Environment asset ID, in short or long form
    :type environment_id: str
    :param registry_name: Name of the registry to create component in
    :type registry_name: str
    :param version_template: Version template
    :type version_template: str
    :return: Environment's full asset ID if successful, else None
    :rtype: Union[str, None]
    """
    try:
        env_name, env_version, env_label, env_registry_name = get_parsed_details_from_asset_uri(
            AssetType.ENVIRONMENT.value, environment_id)
    except Exception as e:
        logger.log_error(e)
        return False

    logger.print(
        f"Env name: {env_name}, version: {env_version}, label: {env_label}, env_registry_name: {env_registry_name}"
    )

    if env_registry_name and env_registry_name not in [PROD_SYSTEM_REGISTRY, registry_name]:
        logger.log_warning(
            f"Dependencies should exist in '{registry_name}' or '{PROD_SYSTEM_REGISTRY}'. "
            f"The URI for environment '{env_name}' references registry '{env_registry_name}', "
            "and publishing will fail if the release process does not have read access to it."
        )

    registry_name = env_registry_name or registry_name

    if env_label:
        if env_label == LATEST_LABEL:
            # TODO: Use a more direct approach like this, when supported by Azure CLI:
            # az ml environment show --name sklearn-1.1-ubuntu20.04-py38-cpu --registry-name azureml --label latest
            versions = get_asset_versions(AssetType.ENVIRONMENT.value, env_name, registry_name)
            if versions:
                # List is returned with the latest version at the beginning
                env_version = versions[0]
            else:
                logger.log_error(f"Unable to retrieve versions for env {env_name}")
                return False
        else:
            # TODO: Add fetching env from other labels
            # https://github.com/Azure/azureml-assets/issues/415
            logger.log_error(f"Creating a component with env label {env_label} is not supported")
            return False

    env = None
    # Get environment
    versions_to_try = [env_version]
    if version_template:
        versions_to_try.append(util.apply_version_template(env_version, version_template))
    for version in versions_to_try:
        if (env := get_asset_details(
            AssetType.ENVIRONMENT.value, env_name, version, registry_name
        )) is not None:
            return env['id']

    logger.log_error(f"Environment {env_name} not found in {registry_name}; tried version(s) {versions_to_try}")
    return None


def validate_update_component(
    spec_path: Path,
    registry_name: str,
    version_template: str = None
) -> bool:
    """Validate and update component spec.

    :param spec_path: Path of loaded component
    :type spec_path: Path
    :param registry_name: name of the registry to create component in
    :type registry_name: str
    :param version_template: version template
    :type version_template: str
    :return: True for successful validation and update
    :rtype: bool
    """
    with open(spec_path) as f:
        try:
            component_dict = YAML().load(f)
        except Exception:
            logger.log_error(f"Error in loading component spec at {spec_path}")
            return False

    component_name = component_dict['name']
    logger.print(f"Preparing component {component_name}")

    # Handle command and parallel components
    if 'environment' in component_dict:
        # Command component
        obj_with_env = component_dict
    elif 'task' in component_dict and 'environment' in component_dict['task']:
        # Parallel component
        obj_with_env = component_dict['task']
    else:
        logger.log_error(f"Environment reference not found in {component_name}")
        return False

    # Update environment reference
    current_env_id = obj_with_env['environment']
    new_env_id = get_environment_asset_id(current_env_id, registry_name, version_template)
    if new_env_id is not None:
        if current_env_id != new_env_id:
            logger.print(f"Updating environment to {new_env_id}")
            obj_with_env['environment'] = new_env_id
        else:
            logger.print(f"Existing environment reference {current_env_id} is valid")
    else:
        return False

    # Update spec file
    try:
        util.dump_yaml(component_dict, spec_path)
    except Exception:
        logger.log_error(f"Component update failed for asset spec path: {spec_path}")
        return False
    return True


def run_command(cmd: List[str]) -> CompletedProcess:
    """Run the command for and return result."""
    result = run(cmd, capture_output=True, encoding=sys.stdout.encoding, errors="ignore")
    return result


def asset_create_command(
    asset_type: str,
    asset_path: str,
    registry_name: str,
    version: str,
    debug: bool = None,
) -> List[str]:
    """Assemble the az cli command."""
    cmd = [
        shutil.which("az"), "ml", asset_type, "create",
        "--file", asset_path,
        "--registry-name", registry_name,
        "--version", version,
    ]
    if debug:
        cmd.append("--debug")
    return cmd


def create_asset_cli(
    asset: AssetConfig,
    registry_name: str,
    version: str,
    debug: bool = None
) -> bool:
    """Create asset in registry."""
    cmd = asset_create_command(
        asset.type.value, str(asset.spec_with_path),
        registry_name, version, debug
    )

    # Run command
    result = run_command(cmd)
    if debug:
        # Capture and redact output
        logger.print(f"Executed: {cmd}")
        redacted_output = sanitize_output(result.stdout)
        if redacted_output:
            logger.print(f"STDOUT: {redacted_output}")

    if result.returncode != 0:
        redacted_err = sanitize_output(result.stderr)
        logger.log_error(f"Error creating {asset.type.value} {asset.name}: {redacted_err}")
        return False
    return True


def get_asset_versions(
    asset_type: str,
    asset_name: str,
    registry_name: str,
) -> List[str]:
    """Get asset versions from registry."""
    cmd = [
        "az", "ml", asset_type, "list",
        "--name", asset_name,
        "--registry-name", registry_name,
    ]
    result = run_command(cmd)
    if result.returncode != 0:
        logger.log_error(f"Failed to list assets: {result.stderr}")
        return []
    return [a['version'] for a in json.loads(result.stdout)]


def get_asset_details(
    asset_type: str,
    asset_name: str,
    asset_version: str,
    registry_name: str,
) -> Dict:
    """Get asset details."""
    logger.print(f"Getting asset details for {asset_type} {asset_name} {asset_version} in {registry_name}")
    cmd = [
        "az", "ml", asset_type, "show",
        "--name", asset_name,
        "--version", asset_version,
        "--registry-name", registry_name,
    ]
    result = run_command(cmd)
    if result.returncode != 0:
        if "Could not find asset" not in result.stderr:
            # Don't show the error if it's expected for new assets
            logger.log_error(f"Failed to get asset details: {result.stderr}")
        return None
    return json.loads(result.stdout)


def get_parsed_details_from_asset_uri(asset_type: str, asset_uri: str) -> Tuple[str, str, str, str]:
    """Validate asset URI and return parsed details. Exception is raised for an invalid URI.

    :param asset_type: Valid values are component, environment and model
    :type asset_type: str
    :param asset_uri: A workspace or registry asset URI to parse
    :type asset_uri: str
    :return:
        A tuple with asset `name`, `version`, `label`, and `registry_name` in order.
        `label` and `registry_name` will be None for workspace URI.
    :rtype: Tuple
    """
    REGISTRY_ASSET_PATTERN = re.compile(REGISTRY_ASSET_TEMPLATE.substitute(
                                        asset_type=pluralize_asset_type(asset_type)))
    asset_registry_name = None
    if (match := REGISTRY_ASSET_PATTERN.match(asset_uri)) is not None:
        asset_registry_name, asset_name, asset_version, asset_label = match.groups()
    elif (match := WORKSPACE_ASSET_PATTERN.match(asset_uri)) is not None:
        asset_name, asset_version, asset_label = match.groups()
    else:
        raise Exception(f"{asset_uri} doesn't match workspace or registry pattern.")
    return asset_name, asset_version, asset_label, asset_registry_name


def stringify_dictionary(dictionary: Dict):
    """Convert the type of values to string."""
    new_dict = {}
    for name, value in dictionary.items():
        new_dict[name] = json.dumps(value) if isinstance(value, dict) else str(value)
    return new_dict


def update_asset_metadata(asset: AssetConfig, ml_client: MLClient, allow_no_op_update: bool = False):
    """Update the mutable metadata of asset."""
    if asset.type == AssetType.MODEL:
        model_name = asset.name
        model_version = asset.version
        spec_path = asset.spec_with_path
        model_config = asset.extra_config_as_object()

        tags_to_update = None
        try:
            with open(spec_path) as f:
                model_spec = YAML().load(f)
                tags = model_spec.get("tags", {})
                properties = model_spec.get("properties", {})

                tags = {k: resolve_from_file_for_asset(model_config, v) for k, v in tags.items()}

                # convert tags, properties value to string
                tags = stringify_dictionary(tags)
                properties = stringify_dictionary(properties)
                tags_to_update = {"replace": tags}
                properties_to_update = {"add": properties}
        except Exception as e:
            logger.log_error(f"Failed to get tags for model {model_name}: {e}")

        update_model_metadata(
            model_name=model_name,
            model_version=model_version,
            update=AssetVersionUpdate(
                versions=[model_version],
                tags=tags_to_update,
                properties=properties_to_update,
                description=model_config.description
            ),
            ml_client=ml_client,
            allow_no_op_update=allow_no_op_update,
        )
    else:
        logger.print(f"Skipping metadata update of {asset.name}. Not supported for type {asset.type}")


def create_asset(asset: AssetConfig, registry_name: str, ml_client: MLClient, version_template: str = None,
                 debug: bool = None, copy_updater: CopyUpdater = None) -> bool:
    """Create asset or update model metadata if it already exists.

    Args:
        asset (AssetConfig): Asset config.
        registry_name (str): Registry name.
        ml_client (MLClient): MLClient object.
        version_template (str, optional): Version template. Defaults to None.
        debug (bool, optional): Enable debug logging. Defaults to None.
        copy_updater (CopyUpdater, optional): CopyUpdater object to update files during azcopy. Defaults to None.

    Returns:
        bool: True of successfully create/updated, otherwise False.
    """
    # Apply version template
    version = util.apply_version_template(asset.version, version_template)
    logger.print(f"Creating {asset.name} {version}")

    # Update model metadata, if it exists
    if get_asset_details(asset.type.value, asset.name, asset.version, registry_name):
        logger.print(f"{asset.name} {asset.version} already exists, updating the metadata")
        try:
            update_asset_metadata(asset=asset, ml_client=ml_client, allow_no_op_update=False)
            return True
        except Exception as e:
            logger.log_error(f"Failed to update metadata for {asset.name}:{asset.version} - {e}")
            return False

    # Handle specific asset types
    with TemporaryDirectory() as temp_dir:
        if asset.type == AssetType.COMPONENT:
            # load component and check if environment exists
            component_type = asset.spec_as_object().type
            if component_type == ComponentType.PIPELINE.value:
                if not validate_and_prepare_pipeline_component(asset.spec_with_path, registry_name, version_template):
                    return False
            elif component_type is None or component_type in [ComponentType.COMMAND.value,
                                                              ComponentType.PARALLEL.value]:
                if not validate_update_component(asset.spec_with_path, registry_name, version_template):
                    return False
        elif asset.type == AssetType.MODEL:
            version = asset.version
            model_config = asset.extra_config_as_object()
            if not prepare_model_for_registration(model_config, asset.spec_with_path, Path(temp_dir), ml_client,
                                                  copy_updater):
                logger.log_error("Failed to prepare model")
                return False

        # Create asset
        return create_asset_cli(
            asset=asset,
            version=version,
            registry_name=registry_name,
            debug=debug,
        )
